// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Leonardo Mezzina, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using BoPi.Types;
using BoPi.Common; 

using IPattern = BoPi.Types.IType;

namespace BoPi.Machine
{
	public interface IMatcher
	{
		int Match(IValue v, Env env);
	}

	public class MatcherFactory
	{
		public static int REPEAT = 1;

		public static void PrintElapsed(DateTime t0, DateTime t1)
		{
#if false
			Console.WriteLine("start = {0} end = {1}", t0, t1);
			Console.WriteLine("elapsed for {0} matches: {1}", MatcherFactory.REPEAT, new TimeSpan(t1.Ticks - t0.Ticks).TotalSeconds);

#endif
		}

		public static IMatcher Make(IPattern pattern)
		{
			ArrayList patterns = new ArrayList();
			patterns.Add(pattern);
			return Make(patterns);
		}

		public static IMatcher Make(ArrayList patterns)
#if false
		{ return new NaiveMatcher(patterns); }
#else
		{ return new SmartMatcher(patterns); }
#endif
	}

	public class NaiveMatcher : IMatcher
	{
		private ArrayList patterns;

		private static ITypeChecker typeChecker = new TypeChecker();

		public NaiveMatcher(ArrayList patterns)
		{ this.patterns = patterns; }

		public static bool Match(IPattern pattern, IValue v, Env env)
		{
			if (pattern is Labelled) {
				if (v is LabelValue) {
					Labelled pat = (Labelled) pattern; 
					LabelValue val = (LabelValue) v;
					if (pat.Labels.Includes(new UnionLabel(val.Label)))
						return Match(pat.Content, val.Content, env);
					else
						return false;
				} else
					return false;
			} else if (pattern is Sequence) {
				Sequence pat = (Sequence) pattern;
				if (v is SequenceValue) {
					SequenceValue val = (SequenceValue) v;
					if (Match(pat.Top, val.Top, env) && Match(pat.Tail, val.Tail, env))
						return true;
				}
				if (Match(pat.Top, new VoidValue(), env) && Match(pat.Tail, v, env))
					return true;
				if (Match(pat.Top, v, env) && Match(pat.Tail, new VoidValue(), env))
					return true;
				return false;
			} else if (pattern is Bind) {
				Bind path = (Bind) pattern;
				if (Match(path.Content, v, env)) {
					env.Set(path.Index, v);
					return true;
				} else
					return false;
			} else { // Schema
				return typeChecker.IsSubtype(v.TypeOf(), pattern);
			}
		}

		public int Match(IValue v, Env env)
		{
			int res = -1;
			DateTime t0 = DateTime.Now;
			for (int j = 0; j < MatcherFactory.REPEAT; j++)
			for (int i = 0; i < patterns.Count; i++)
				if (Match((IPattern) patterns[i], v, env)) {
					res = i;
					break;
				}
			DateTime t1 = DateTime.Now;
			MatcherFactory.PrintElapsed(t0, t1);
			return res;
		}
	}

	public class SmartMatcher : IMatcher
	{
		private IBasicMatcher matcher;

		private enum MatchClass { DONE, BIND, LABEL, SEQ, SCHEMA, MIX }

		private class Cell<T>
		{
			private T content;
			private Cell<T> next;

			public Cell(T content, Cell<T> next)
			{
				this.content = content;
				this.next = next;
			}

			public T Content { get { return this.content; } }
			public Cell<T> Next { get { return this.next; } }
		}

		public class NoMatch : Exception
		{ public NoMatch(String message) : base("No match: " + message) { } }

		public SmartMatcher(ArrayList patterns)
		{
			matcher = new Matrix(patterns).Compile();
		}

		private interface IBasicMatcher
		{
			int Match(Env env, Cell<IValue> vm, Cell<IValue> v);
			void Dump();
		}

		private class DoneMatch : IBasicMatcher
		{
			private int row;
			private Cell<int> indexes;

			public DoneMatch(int row, Cell<int> indexes)
			{
				this.row = row;
				this.indexes = indexes;
			}

			public void Dump()
			{ Console.WriteLine("<done row=\"{0}\"/>", row); }

			public int Match(Env env, Cell<IValue> vm, Cell<IValue> v)
			{
				Debug.Assert(v == null, "pattern matching ended?");

				Cell<int> p = indexes;
				Cell<IValue> q = vm;
				while (p != null) {
					Debug.Assert(q != null);
					env.Set(p.Content, q.Content);
					p = p.Next;
					q = q.Next;
				}

				return row;
			}
		}

		private class BindMatch : IBasicMatcher
		{
			private IBasicMatcher cont;

			public BindMatch(IBasicMatcher cont)
			{ this.cont = cont; }

			public void Dump()
			{
				Console.WriteLine("<bind>");
				cont.Dump();
				Console.WriteLine("</bind>");
			}

			public int Match(Env env, Cell<IValue> vm, Cell<IValue> v)
			{
				Debug.Assert(v != null);
				return cont.Match(env, new Cell<IValue>(v.Content, vm), v);
			}
		}

		private class SequenceMatch : IBasicMatcher
		{
			private IBasicMatcher cont;

			public SequenceMatch(IBasicMatcher cont)
			{ this.cont = cont; }

			public void Dump()
			{
				Console.WriteLine("<sequence>");
				cont.Dump();
				Console.WriteLine("</sequence>");
			}

			public int Match(Env env, Cell<IValue> vm, Cell<IValue> v)
			{
				if (v.Content is SequenceValue) {
					SequenceValue vSeq = (SequenceValue) v.Content;
					return cont.Match(env, vm, new Cell<IValue>(vSeq.Top, new Cell<IValue>(vSeq.Tail, v.Next)));
				}
				throw new NoMatch("seq");
			}
		}

		private class BacktrackMatch : IBasicMatcher
		{
			private IBasicMatcher tryBranch;
			private IBasicMatcher catchBranch;

			public BacktrackMatch(IBasicMatcher tryBranch, IBasicMatcher catchBranch)
			{
				this.tryBranch = tryBranch;
				this.catchBranch = catchBranch;
			}

			public void Dump()
			{
				Console.WriteLine("<try>");
				tryBranch.Dump();
				catchBranch.Dump();
				Console.WriteLine("</try>");
			}

			public int Match(Env env, Cell<IValue> vm, Cell<IValue> v)
			{
				try {
					return tryBranch.Match(env, vm, v);
				} catch (NoMatch exc) {
					return catchBranch.Match(env, vm, v);
				}
			}
		}

		private class LabelMatch : IBasicMatcher
		{
			private LabelsSet labels;
			private IBasicMatcher cont;

			public LabelMatch(LabelsSet labels, IBasicMatcher cont)
			{
				this.labels = labels;
				this.cont = cont;
			}

			public void Dump()
			{
				Console.WriteLine("<label labels=\"{0}\">", labels);
				cont.Dump();
				Console.WriteLine("</label>");
			}

			public int Match(Env env, Cell<IValue> vm, Cell<IValue> v)
			{
				if (v.Content is LabelValue) {
					LabelValue lab = (LabelValue) v.Content;
					if (labels.Includes(new UnionLabel(lab.Label)))
						return cont.Match(env, vm, new Cell<IValue>(lab.Content, v.Next));
				}
				throw new NoMatch("label");
			}
		}

		private class SchemaMatch : IBasicMatcher
		{
			private IType schema;
			private IBasicMatcher cont;

			private static ITypeChecker typeChecker = new TypeChecker();

			public SchemaMatch(IType schema, IBasicMatcher cont)
			{
				this.schema = schema;
				this.cont = cont;
			}

			public void Dump()
			{
				Console.WriteLine("<schema type=\"{0}\">", schema);
				cont.Dump();
				Console.WriteLine("</schema>");
			}

			public int Match(Env env, Cell<IValue> vm, Cell<IValue> v)
			{
				Debug.Assert(v != null);
				if (typeChecker.IsSubtype(v.Content.TypeOf(), schema))
					return cont.Match(env, vm, v.Next);
				else
					throw new NoMatch("schema");
			}
		}

		private class Matrix
		{
			private List<Stack<IPattern>> content;
			private List<int> rows;
			private List<Cell<int>> envs;

			public Matrix(ArrayList patterns)
			{
				content = new List<Stack<IPattern>>();
				rows = new List<int>();
				envs = new List<Cell<int>>();

				for (int i = 0; i < patterns.Count; i++)
					rows.Add(i);

				foreach (IPattern pat in patterns) {
					Stack<IPattern> row = new Stack<IPattern>();
					row.Push(pat);
					content.Add(row);
					envs.Add(null);
				}
			}

			private Matrix(List<Stack<IPattern>> content, List<int> rows, List<Cell<int>> envs, int first, int last)
			{
				this.content = content.GetRange(first, last - first + 1);
				this.rows = rows.GetRange(first, last - first + 1);
				this.envs = envs.GetRange(first, last - first + 1);
			}

			private void Expand()
			{
				int i = 0;
				while (i < content.Count) {
					Stack<IPattern> row = content[i];
					if (row.Count == 0)
						break;
	
			  	IPattern pat = row.Peek();
					if (pat is Union) {
						Union u = (Union) pat;
						row.Pop();
						Stack<IPattern> newRow = new Stack<IPattern>(row);
						row.Push(u.Fst);
						newRow.Push(u.Snd);
						content.Insert(i, newRow);
						rows.Insert(i, rows[i]);
						envs.Insert(i, envs[i]);
					} else
						i++;
				}
			}

			private void Split(int n, out Matrix prefix, out Matrix rest)
			{
				if (n < content.Count) {
					prefix = new Matrix(content, rows, envs, 0, n - 1);
					rest = new Matrix(content, rows, envs, n, content.Count - 1);
				} else {
					Debug.Assert(n == content.Count);
					prefix = this;
					rest = null;
				}
			}

			public MatchClass Classify(out Matrix prefix, out Matrix rest)
			{
				Expand();

				if (content[0].Count == 0) {
					prefix = this;
					rest = null;
					return MatchClass.DONE;
				}

				MatchClass state = MatchClass.DONE;
				int n = 0;
				bool done = false;
				while (!done && n < content.Count) {
			  	IPattern pat = content[n].Peek();
			  	switch (state) {
				  	case MatchClass.DONE:
							if (pat is Labelled)
								state = MatchClass.LABEL;
							else if (pat is Sequence)
								state = MatchClass.SEQ;
							else if (pat is Bind)
								state = MatchClass.BIND;
							else
								state = MatchClass.SCHEMA;
							break;
						case MatchClass.SCHEMA:
							if ((pat is Labelled || pat is Sequence || pat is Bind))
								done = true;
							else {
								IType t1 = pat;
								IType t2 = content[0].Peek();
								ITypeChecker typeChecker = new TypeChecker();
								done = !(typeChecker.IsSubtype(t1, t2) && typeChecker.IsSubtype(t2, t1));
							}
					  	break;
						case MatchClass.LABEL:
							if (pat is Labelled) {
								LabelsSet set1 = ((Labelled) pat).Labels;
								LabelsSet set2 = ((Labelled) (content[0]).Peek()).Labels;
								done = !set1.Includes(set2) || !set2.Includes(set1);
							} else
								done = true;
							break;
						case MatchClass.SEQ:
							done = !(pat is Sequence);
					  	break;
						case MatchClass.BIND:
							done = !(pat is Bind);
							break;
						default:
							Debug.Assert(false);
							break;
					}

					if (!done) n++;
				}

				Debug.Assert(n > 0, "n > 0");
				Split(n, out prefix, out rest);
				return state;
			}

			private IBasicMatcher CompileBind()
			{
				for (int i = 0; i < content.Count; i++) {
					Stack<IPattern> row = content[i];
					Debug.Assert(row.Peek() is Bind);
					Bind var = (Bind) row.Pop();
					envs[i] = new Cell<int>(var.Index, (Cell<int>) envs[i]);
					row.Push(var.Content);
				}
				return new BindMatch(Compile());
			}

			private IBasicMatcher CompileSeq()
			{
				for (int i = 0; i < content.Count; i++) {
					Stack<IPattern> row = content[i];
					Debug.Assert(row.Peek() is Sequence);
					Sequence seq = (Sequence) row.Pop();
					row.Push(seq.Tail);
					row.Push(seq.Top);
				}
				return new SequenceMatch(Compile());
			}

			private IBasicMatcher CompileLabel()
			{
				Debug.Assert(content[0].Peek() is Labelled);
				Labelled pat = (Labelled) content[0].Peek();
				for (int i = 0; i < content.Count; i++) {
					Stack<IPattern> row = content[i];
					Debug.Assert(row.Peek() is Labelled);
					Labelled rowPat = (Labelled) (row.Pop());
					row.Push(rowPat.Content);
				}
				return new LabelMatch(pat.Labels, Compile());
			}

			private IBasicMatcher CompileSchema()
			{
				IType schema = content[0].Peek();
				for (int i = 0; i < content.Count; i++)
					content[i].Pop();
				return new SchemaMatch(schema, Compile());
			}

			public IBasicMatcher Compile()
			{
				Matrix prefix;
				Matrix rest;
				MatchClass cls = Classify(out prefix, out rest);
				IBasicMatcher res = null;
				switch (cls) {
					case MatchClass.DONE:
						return new DoneMatch(rows[0], envs[0]);
						break;
					case MatchClass.BIND:
						res = prefix.CompileBind();
						break;
					case MatchClass.SEQ:
						res = prefix.CompileSeq();
						break;
					case MatchClass.LABEL:
						res = prefix.CompileLabel();
						break;
					case MatchClass.SCHEMA:
						res = prefix.CompileSchema();
						break;
					default:
						Debug.Assert(false);
						break;
				}

				Debug.Assert(res != null, "res != null");
				if (rest != null)
					res = new BacktrackMatch(res, rest.Compile());

				return res;
			}
		}

		public int Match(IValue v, Env env)
		{
#if false
			Console.WriteLine("====");
			matcher.Dump();
			Console.WriteLine("====");
#endif
			int res = 0;
			DateTime t0 = DateTime.Now;
			for (int i = 0; i < MatcherFactory.REPEAT; i++)
				res = matcher.Match(env, null, new Cell<IValue>(v, null));
			DateTime t1 = DateTime.Now;
			MatcherFactory.PrintElapsed(t0, t1);
			return res;
		}
	}
}
