// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Leonardo Mezzina, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections.Generic;
using System.Collections;
using BoPi.Common;
using System.Diagnostics;

namespace BoPi.Types
{
  public class FirstTypeVisitor : ITypeVisitor
  {
    private IList<IType> visited = new List<IType>();
    private IList<ILabelsSet> first = new List<ILabelsSet>();

    public IList<ILabelsSet> First { get { return first; } }

    public void VisitBottom(Bottom t)
    {  }
    
    public void VisitVoid(Void t)
    { first.Add(new VoidLabel()); }

    public void VisitUnion(Union t)
    {
      t.Fst.Accept(this);
      t.Snd.Accept(this);
    }

    public void VisitConstantTypeName(ConstantTypeName t)
    {
      if (!visited.Contains(t))
      {
        visited.Add(t);
        t.Entry.Type.Accept(this);
      }
    }

    public void VisitSequence(Sequence t)
    { t.Top.Accept(this); }

    public void VisitLabelled(Labelled t)
    { first.Add(t.Labels); }

    public void VisitIntLiteral(IntLiteral t)
    { first.Add(new IntLiteralLabel(t.Val)); }

    public void VisitIntType(IntType t)
    { first.Add(new IntLabel()); }

    public void VisitStringLiteral(StringLiteral t)
    { first.Add(new StringLiteralLabel(t.Val)); }

    public void VisitStringType(StringType t)
    { first.Add(new StringLabel()); }

    public void VisitChan(Chan t)
    { first.Add(new ChanLabel(t.Capability)); }

    public void VisitErrorType(ErrorType t)
    {}

    public void VisitBind(Bind t)
    { t.Content.Accept(this); }
  }
    
  public class DeterminednessChecker
  {
    private ICollection<IType> ndet = new List<IType>();
    private ICollection<IType> cdet = new List<IType>();
    private IDictionary<IType, IList<ILabelsSet>> dict = new Dictionary<IType, IList<ILabelsSet>>();
    private EmptynessChecker ec = new EmptynessChecker();
   	    
    public IList<ILabelsSet> GetFirst(IType t)
    {
      IList<ILabelsSet> first;
      if (!dict.TryGetValue(t, out first))
      {	      
        FirstTypeVisitor visitor = new FirstTypeVisitor();
        t.Accept(visitor);
	      first = visitor.First;
      }
      return first;
    }

    /// <summary>
    /// Check if the type is determined. A type is determined if every union
    /// starts with a different label. <c>a[int]+a[string]</c> is not
    /// determined.
    /// </summary>
    /// <param name="s">the type</param>
    /// <returns>true if the type is determined, false otherwise</returns>
    public bool IsDetermined(IType s)
    {
      if (cdet.Contains(s)) return true;
      else if (ndet.Contains(s)) return false;
      bool res = IsDetermined(s, new ArraySet()); 
      if (res) cdet.Add(s);
      else ndet.Add(s);
      return res;
    }

    private bool IsDetermined(IType s, ISet det)
    {
      if (s.IsBottom() || s.IsBaseType() || s.IsVoid() || s.IsErrorType())
        return true;
      else  if (s.IsChan())  
        return IsDetermined(s.AsChan().Carried, det);
      else if (s.IsLabelled())
        return IsDetermined(s.AsLabelled().Content, det);
      else if (s.IsSequence()) 
        return IsDetermined(s.AsSequence().Top, det) && IsDetermined(s.AsSequence().Tail, det);
      else if (s.IsUnion())
      {
        Union u = s.AsUnion();
        if (ec.IsEmpty(u.Fst) || ec.IsEmpty(u.Snd))
          return true;
        else
        {
          bool res = true;
          IList<ILabelsSet> firstOfFst = GetFirst(u.Fst);
          IList<ILabelsSet> firstOfSnd = GetFirst(u.Snd);
          for (int i = 0; i < firstOfFst.Count && res; i++)
            for (int j = 0; j < firstOfSnd.Count && res; j++)
              if (firstOfFst[i] is LabelsSet && firstOfSnd[j] is LabelsSet)
                res &= ((LabelsSet)firstOfFst[i]).Intersect((LabelsSet)firstOfSnd[j]).IsEmpty();
          return res && IsDetermined(u.Fst, det) && IsDetermined(u.Snd, det);
        }
      }
      else if (s.IsConstantTypeName()) 
      {
        if (det.Contains(s)) return true;      
        det.Add(s);
        return IsDetermined(s.AsConstantTypeName().Entry.Type, det);
      }
      else throw new ApplicationException("IsDetermined: Unhandled case " + s);
    }  
  }
}
